# -*- coding: utf-8 -*-

"""This code is a part of Hydra Toolkit

.. module:: hydratk.extensions.datagen.translation.en
   :platform: Unix
   :synopsis: English language translation for Datagen extension
.. moduleauthor:: Petr Czaderna <pc@hydratk.org>

"""

language = {
  'name' : 'English',
  'ISO-639-1' : 'en'
}

msg = {
    'datagen_received_cmd'           : ["Received command: '{0}'"],     
    'datagen_asn1_import_spec'       : ["Import ASN.1 specification: '{0}'"],
    'datagen_asn1_spec_imported'     : ["ASN.1 specification imported"],
    'datagen_asn1_encode'            : ["Encoding input text file: '{0}'"],
    'datagen_asn1_encoded'           : ["Encoded to output binary file: '{0}'"],
    'datagen_asn1_decode'            : ["Decoding input binary file: '{0}'"],
    'datagen_asn1_decoded'           : ["Decoded to output text file: '{0}'"],
    'datagen_jsongen_import_spec'    : ["Importing JSON specification: '{0}'"],
    'datagen_jsongen_spec_imported'  : ["JSON specification imported"],
    'datagen_jsongen_write_sample'   : ["Creating sample JSON document"],
    'datagen_jsongen_sample_written' : ["Sample written to file: '{0}'"],
    'datagen_xmlgen_import_spec'     : ["Importing XML specification: '{0}'"],
    'datagen_xmlgen_spec_imported'   : ["XML specification imported"],
    'datagen_xmlgen_write_sample'    : ["Creating sample XML document"],
    'datagen_xmlgen_sample_written'  : ["Sample written to file: '{0}'"]                      
}
