# -*- coding: utf-8 -*-

"""This code is a part of Hydra Toolkit

.. module:: hydratk.extensions.datagen.translation.cs
   :platform: Unix
   :synopsis: Czech language translation for Datagen extension
.. moduleauthor:: Petr Rašek <bowman@hydratk.org>

"""

language = {
  'name' : 'Čeština',
  'ISO-639-1' : 'cs'
} 


msg = {
    'datagen_received_cmd'           : ["Obdržen příkaz: '{0}'"], 
    'datagen_asn1_import_spec'       : ["Importuji ASN.1 specifikaci: '{0}'"],
    'datagen_asn1_spec_imported'     : ["ASN.1 specifikace importována"],
    'datagen_asn1_encode'            : ["Enkóduji vstupní textový soubor: '{0}'"],
    'datagen_asn1_encoded'           : ["Enkódováno do binárního souboru: '{0}'"],
    'datagen_asn1_decode'            : ["Dekóduji vstupní binární soubor: '{0}'"],
    'datagen_asn1_decoded'           : ["Dekódováno do výstupního textového souboru: '{0}'"],
    'datagen_jsongen_import_spec'    : ["Importuji JSON specifikaci: '{0}'"],
    'datagen_jsongen_spec_imported'  : ["JSON specifikace importována"],
    'datagen_jsongen_write_sample'   : ["Vytvářím vzorový JSON dokument"],
    'datagen_jsongen_sample_written' : ["Vzorek zapsán do souboru: '{0}'"],
    'datagen_xmlgen_import_spec'     : ["Importuji XML specifikaci: '{0}'"],
    'datagen_xmlgen_spec_imported'   : ["XML specifikace importována"],
    'datagen_xmlgen_write_sample'    : ["Vytvářím vzorový XML dokument"],
    'datagen_xmlgen_sample_written'  : ["Vzorek zapsán do souboru: '{0}'"]               
}